"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.spreadMatcher = void 0;
const lodash_1 = require("lodash");
const ethers_1 = require("ethers");
const spreadMatcher = (originalData) => {
    const data = lodash_1.cloneDeep(originalData);
    const matches = [];
    let i = 0;
    while (true) {
        if (i > 10000) {
            throw new Error(`
              Matching executed more than 10000 - it may mean, that something broke
              data: ${JSON.stringify(data, null, 2)}
              matches: not displayed, because of potential length
            `);
        }
        const allGroupsHaveVolume = [
            data.entityGroups[0].some((v) => v.volume.gt(0)),
            data.entityGroups[1].some((v) => v.volume.gt(0))
        ].every(Boolean);
        if (!allGroupsHaveVolume) {
            break;
        }
        const roundMatches = matchRound(getRoundInput(data));
        roundMatches.forEach((match) => {
            const entity1 = data.entityGroups[0].find((c) => c.id === match.entities[0]);
            const entity2 = data.entityGroups[1].find((g) => g.id === match.entities[1]);
            entity1.volume = entity1.volume.sub(match.volume);
            entity2.volume = entity2.volume.sub(match.volume);
        });
        matches.push(roundMatches);
        i += 1;
    }
    return {
        matches: sumMatches(matches.flat()),
        leftoverEntities: data.entityGroups
    };
};
exports.spreadMatcher = spreadMatcher;
const sumMatches = (matches) => {
    return Object.values(lodash_1.groupBy(matches, (m) => `${m.entities[0]}${m.entities[1]}`)).map((matches) => ({
        entities: matches[0].entities,
        volume: bigNumSum(matches.map((m) => m.volume))
    }));
};
const computeSafeDistribution = (entities1, entities2) => {
    const compute = (entities, otherEntitiesLength) => entities.map((entity) => ({
        entityId: entity.id,
        volume: entity.volume,
        distributedVolume: entity.volume.div(otherEntitiesLength)
    }));
    return bigNumMinBy([...compute(entities1, entities2.length), ...compute(entities2, entities1.length)], (e) => e.distributedVolume);
};
const getCompetingEntities = (routes, entityId) => {
    return routes
        .filter((route) => route.includes(entityId))
        .flat()
        .filter((id) => id !== entityId);
};
const matchRound = (input) => {
    const entityToDistribute = computeSafeDistribution(input.entityGroups[0], input.entityGroups[1]);
    if (entityToDistribute.distributedVolume.eq(0)) {
        const competingEntities = getCompetingEntities(input.routes, entityToDistribute.entityId);
        const entitiesCountToUse = entityToDistribute.volume.toNumber();
        const isEntityFirst = input.entityGroups[0].find((e) => e.id === entityToDistribute.entityId);
        return competingEntities.slice(0, entitiesCountToUse).map((connectedEntity) => ({
            entities: isEntityFirst
                ? [entityToDistribute.entityId, connectedEntity]
                : [connectedEntity, entityToDistribute.entityId],
            volume: ethers_1.BigNumber.from(1)
        }));
    }
    return input.routes.map((route) => ({
        entities: route,
        volume: entityToDistribute.distributedVolume
    }));
};
const getRoundInput = (data) => {
    const entity1round = getFirstUsableGroup(data.priority, data.entityGroups[0]);
    const routes = entity1round.flatMap((entity1) => {
        const preferredEntity2Group = getFirstUsableGroup(entity1.priority, data.entityGroups[1]);
        return preferredEntity2Group.map((entity2) => [entity1.id, entity2.id]);
    });
    return {
        routes,
        entityGroups: [
            data.entityGroups[0].filter((entity) => routes.some((e) => e[0] === entity.id)),
            data.entityGroups[1].filter((entity) => routes.some((e) => e[1] === entity.id))
        ]
    };
};
const getFirstUsableGroup = (groups, data) => {
    const groupsWithVolume = groups
        .map((group) => {
        const withVolume = group.map((entry) => ({
            ...entry,
            volume: data.find((c) => c.id === entry.id)?.volume ?? ethers_1.BigNumber.from(0)
        }));
        return withVolume.filter((entry) => entry.volume.gt(0));
    })
        .filter((group) => group.length > 0);
    return groupsWithVolume[0] ?? [];
};
const bigNumSum = (bigNumbers) => bigNumbers.reduce((sum, v) => sum.add(v), ethers_1.BigNumber.from(0));
const bigNumMinBy = (collection, cb) => {
    return collection.reduce((smallest, current) => {
        return cb(current).lt(cb(smallest)) ? current : smallest;
    }, collection[0]);
};
//# sourceMappingURL=spreadMatcher.js.map